/**
    \file sigslot_demo.cpp
    \author Kurt Stege

    Brief demonstration of the Sigslot concept.

    These examples are used to develop and test the Sigslot concept
    within its first weeks of development.

    History:

    (ks) = Kurt Stege, kurt-stege@online.de

    2005-07-18 (ks):    File created.
    2005-09-08 (ks):    Current version.
*/

#include <stdio.h>

#ifdef _MSC_VER
#pragma warning(disable:4786) // warning C4786: identifier was truncated to '255' characters in the debug information
#pragma warning(disable:4355) // warning C4355: 'this' : used in base member initializer list
#endif

#include "Sigslot/TSignature.hpp"
#include "Sigslot/TJunctionBox.hpp"
#include "Sigslot/TSlot.hpp"
#include "Sigslot/CThread.hpp"
#include "Base/CThread.hpp" // for sleep()


class CDemo
{
    public:

    // slot input(double value);
    // NSigslot::slot input(double value);
    void input(double value);

    // signal output(double value);
    // NSigslot::signal output(double value);
    NSigslot::TSignature1<double>::CSignal output;
};

void CDemo::input(double value)
{
    output(value);
}


typedef NSigslot::TSignature1<double>   tValueSignature;

class CPrint
{
    public:

    void SLOT_print (double value)
    {
        printf ("component CPrint(%p): slot print called with value %f.\n", this, value);
    }
};


class CMultiply
{
    public:

    tValueSignature::CSignal SIGNAL_mult2;
    tValueSignature::CSignal SIGNAL_mult3;
    tValueSignature::CSignal SIGNAL_mult4;

    void SLOT_input (double value)
    {
        SIGNAL_mult2 (value * 2.0);
        SIGNAL_mult3 (value * 3.0);
        SIGNAL_mult4 (value * 4.0);
    }
};

class CAverage
{
    public:

    tValueSignature::CSignal SIGNAL_newAverage;

    void SLOT_reset()
    {
        m_sum = 0.0;
        m_count = 0;
    }

    void SLOT_processValue (double value)
    {
        m_count++;
        m_sum += value;
        SIGNAL_newAverage (m_sum/m_count);
    }

    CAverage()
    :
        m_sum(0.0),
        m_count(0)
    {
    }

    ~CAverage()
    {
    }

    private:

    double m_sum;
    int m_count;
};


class CInput
{
    public:

    tValueSignature::CSignal SIGNAL_newValue;

    void processInput (double inputValue)
    {
        SIGNAL_newValue (inputValue);
    }
};



/*
    Note from Kurt for Kurt:

    Cable-Template ...
 (v)... to an ordinary slot of any component class (const, non-const, or static)
    ... to any global or static function
    ... to a signal
    ... to a compressed slot of any class (receives a const CMessage &msg instead of a parameter list)
 v  ... to an operator()(parameterlist) of any class
 v  ... to an operator()(const CMessage &msg) of any class
    ... ...
*/




/*!
    This is a way supposed to be used by the code generation tools
    to create a connection, when a signal is to be received by
    exactly one slot. Please note that this is probably the most
    typical connection.
*/
class CCableTo_SLOT_print_ofCPrint : public tValueSignature::CSignal
{
    private:

    static void cableFunction (void *context, const tValueSignature::CMessage &msg)
    {
        CPrint *receiver = static_cast<CPrint *>(context);
        receiver->SLOT_print (msg.m_par1);
    }

    public:

    CCableTo_SLOT_print_ofCPrint(const CPrint &receiver)
    :
        CSignal(cableFunction, static_cast<const void *>(&receiver))
    {
    }
};


/*!
    This is a way supposed to be used by the code generation tools
    to create a connection, when a signal is to be received by
    exactly one slot. Please note that this is probably the most
    typical connection.
*/
template <class TemplateParameterReceiver>
class TCableTo_SLOT_processValue : public tValueSignature::CSignal
{
    public:

    typedef TemplateParameterReceiver tReceiver;

    TCableTo_SLOT_processValue(const tReceiver &receiver)
    :
        CSignal(cableFunction, static_cast<const void *>(&receiver))
    {
    }

    private:

    static void cableFunction (void *context, const tValueSignature::CMessage &msg)
    {
        tReceiver *receiver = static_cast<tReceiver *>(context);
        receiver->SLOT_processValue (msg.m_par1);
    }
};


/*!
    This is a way supposed to be used by the code generation tools
    to create a connection, when a signal is to be received by
    exactly one slot. Please note that this is probably the most
    typical connection.
*/
static void cableFunctionTo_SLOT_input_ofCMultiply (void *context, const tValueSignature::CMessage &msg)
{
    CMultiply *receiver = static_cast<CMultiply *>(context);
    receiver->SLOT_input (msg.m_par1);
}



class CCalculator
{
    private:

    CInput      m_input;
    CMultiply   m_mult;
    CAverage    m_average1;
    CAverage    m_average2;
    CPrint      m_print;

    NSigslot::TJunctionBox<tValueSignature>   m_box1;
    NSigslot::TJunctionBox<tValueSignature>   m_box2;
    NSigslot::TJunctionBox<tValueSignature>   m_box3;

    public:

    CCalculator();
    void createConnections();
    void processAValue(double value);
    static void cableFunction1(void *context, const tValueSignature::CMessage &msg);
};

void CCalculator::createConnections ()
{
#if 1
    /*
        These are several ways to create connections. All these
        ways are supposed to be (possibly) used be some code
        generation tool. It is possible to use these technics
        for hand written code, but there are other ways to
        manually create  connections that are easier to write
        (alas paid by some run-time overhead).
    */

    m_input.SIGNAL_newValue = tValueSignature::CSignal(cableFunction1, this);
    m_average2.SIGNAL_newAverage = tValueSignature::CSignal(cableFunctionTo_SLOT_input_ofCMultiply, static_cast<const void *>(&m_mult));
    m_mult.SIGNAL_mult2 = CCableTo_SLOT_print_ofCPrint(m_print);
    m_mult.SIGNAL_mult3 = TCableTo_SLOT_processValue<CAverage>(m_average1);
    m_average1.SIGNAL_newAverage = CCableTo_SLOT_print_ofCPrint(m_print);

#else
    // Connections are done within the constructor
    // Error: This strategy leads to memory leaks!
    // Provide some special new with garbage collection at least at the end of the program.
    new NSigslot::TJunctionBox<tValueSignature> (m_input.SIGNAL_newValue, CCableTo_SLOT_print_ofCPrint(m_print));
    new NSigslot::TJunctionBox<tValueSignature> (m_input.SIGNAL_newValue, TCableTo_SLOT_processValue<CAverage>(m_average1));
    new NSigslot::TJunctionBox<tValueSignature> (m_average1.SIGNAL_newAverage, CCableTo_SLOT_print_ofCPrint(m_print));
#endif
}

CCalculator::CCalculator()
:
    m_box1(m_input.SIGNAL_newValue, CCableTo_SLOT_print_ofCPrint(m_print)),
    m_box2(m_input.SIGNAL_newValue, TCableTo_SLOT_processValue<CAverage>(m_average1)),
    m_box3(m_average1.SIGNAL_newAverage, CCableTo_SLOT_print_ofCPrint(m_print))
{
}

void CCalculator::processAValue(double value)
{
    m_input.processInput (value);
}

void CCalculator::cableFunction1(void *context, const tValueSignature::CMessage &msg)
{
    CCalculator *this_ = static_cast<CCalculator *>(context);
    this_->m_print.SLOT_print (msg.m_par1);
    this_->m_average1.SLOT_processValue (msg.m_par1);
    this_->m_average2.SLOT_processValue (msg.m_par1);
}


// ******************************************************************
// ******************************************************************
// ******************************************************************


template<typename TemplateParameterSlot, typename TemplateParameterMemberFunctionPointer>
class TProblematicCableToSlot : public /* typename */ TemplateParameterSlot::tSignature::CSignal
{
    public:

    typedef TemplateParameterSlot tSlot;
    typedef TemplateParameterMemberFunctionPointer tMemberFunctionPointer;
    typedef typename tSlot::tSignature tSignature;
    typedef typename tSlot::tReceiver tReceiver;

    TProblematicCableToSlot(tReceiver &receiver, tMemberFunctionPointer slot)
    :
        CSignal(CHelper_nonconst::cableFunction, new CHelper_nonconst(receiver, slot))
        // TODO: Error: This new gets never freed and is a memory leak! This is why is "problematic" in the name.
    {
        printf ("Constructor of TProblematicCableToSlot (nonconst) is called.\n");
    }

    TProblematicCableToSlot(const tReceiver &receiver, tMemberFunctionPointer slot)
    :
        CSignal(CHelper_const::cableFunction, new CHelper_const(receiver, slot))
        // TODO: Error: This new gets never freed and is a memory leak! This is why is "problematic" in the name.
    {
        printf ("Constructor of TProblematicCableToSlot (const) is called.\n");
    }

    private:

    class CHelper_nonconst
    {
        public:

        tReceiver &m_receiver;
        tMemberFunctionPointer m_slot;

        CHelper_nonconst(tReceiver &receiver, tMemberFunctionPointer slot)
        :
            m_receiver(receiver),
            m_slot(slot)
        {
        }

        static void cableFunction (void *context, const typename tSignature::CMessage &msg)
        {
            const CHelper_nonconst *this_ = static_cast<const CHelper_nonconst *>(context);
            tSlot::callSlot (this_->m_receiver, this_->m_slot, msg);
        }
    };

    class CHelper_const
    {
        public:

        const tReceiver &m_receiver;
        tMemberFunctionPointer m_slot;

        CHelper_const(const tReceiver &receiver, tMemberFunctionPointer slot)
        :
            m_receiver(receiver),
            m_slot(slot)
        {
        }

        static void cableFunction (void *context, const typename tSignature::CMessage &msg)
        {
            const CHelper_const *this_ = static_cast<const CHelper_const *>(context);
            tSlot::callSlot (this_->m_receiver, this_->m_slot, msg);
        }
    };
};



template<typename TemplateParameterSlot>
class TIncorrectCableToSlot : public /* typename */ TemplateParameterSlot::tSignature::CSignal
{
    public:

    typedef TemplateParameterSlot tSlot;
    typedef typename tSlot::tSignature tSignature;
    typedef typename tSlot::tReceiver tReceiver;
    typedef typename tSlot::memberFunctionPointer_nonconst memberFunctionPointer_nonconst;
    typedef typename tSlot::memberFunctionPointer_const memberFunctionPointer_const;
    typedef typename tSlot::memberFunctionPointer_static memberFunctionPointer_static;

    TIncorrectCableToSlot(tReceiver &receiver, memberFunctionPointer_nonconst slot)
    :
        CSignal(CHelper_nonconst::cableFunction, new CHelper_nonconst(receiver, slot))
        // TODO: Error: This new gets never freed and is a memory leak! This is why is "incorrect" in the name.
    {
        printf ("Constructor of TIncorrectCableToSlot (nonconst variant) is called.\n");
    }

    TIncorrectCableToSlot(const tReceiver &receiver, memberFunctionPointer_const slot)
    :
        CSignal(CHelper_const::cableFunction, new CHelper_const(receiver, slot))
        // TODO: Error: This new gets never freed and is a memory leak! This is why is "incorrect" in the name.
    {
        printf ("Constructor of TIncorrectCableToSlot (const variant) is called.\n");
    }

    TIncorrectCableToSlot(const tReceiver &receiver, memberFunctionPointer_static slot)
    :
        CSignal(cableFunction_static, slot)
    {
        printf ("Constructor of TIncorrectCableToSlot (static variant) is called This variant is correct.\n");
    }

    private:

    class CHelper_const
    {
        public:

        const tReceiver &m_receiver;
        memberFunctionPointer_const m_slot;

        CHelper_const (const tReceiver &receiver, memberFunctionPointer_const slot)
        :
            m_receiver(receiver),
            m_slot(slot)
        {
        }

        static void cableFunction (void *context, const typename tSignature::CMessage &msg)
        {
            const CHelper_const *this_ = static_cast<const CHelper_const *>(context);
            tSlot::callSlot (this_->m_receiver, this_->m_slot, msg);
        }
    };

    class CHelper_nonconst
    {
        public:

        tReceiver &m_receiver;
        memberFunctionPointer_nonconst m_slot;

        CHelper_nonconst (tReceiver &receiver, memberFunctionPointer_nonconst slot)
        :
            m_receiver(receiver),
            m_slot(slot)
        {
        }

        static void cableFunction (void *context, const typename tSignature::CMessage &msg)
        {
            const CHelper_nonconst *this_ = static_cast<const CHelper_nonconst *>(context);
            tSlot::callSlot (this_->m_receiver, this_->m_slot, msg);
        }
    };

    static void cableFunction_static (void *context, const typename tSignature::CMessage &msg)
    {
        memberFunctionPointer_static slot_static = static_cast<memberFunctionPointer_static>(context);
        tSlot::callSlot (slot_static, msg);
    }
};


template<typename TemplateParameterSlot>
class TDangerousCableToSlot : public /* typename */ TemplateParameterSlot::tSignature::CSignal
{
    public:

    typedef TemplateParameterSlot tSlot;
    typedef typename tSlot::tSignature tSignature;
    typedef typename tSlot::tReceiver tReceiver;
    typedef typename tSlot::memberFunctionPointer_nonconst memberFunctionPointer_nonconst;
    typedef typename tSlot::memberFunctionPointer_const memberFunctionPointer_const;
    typedef typename tSlot::memberFunctionPointer_static memberFunctionPointer_static;

    TDangerousCableToSlot(const tReceiver &receiver, memberFunctionPointer_const slot)
    :
        m_receiver(&receiver),
        m_slot_const(slot),
        m_slot_nonconst(0),
        CSignal(cableFunction_const, this)
        /*
            TODO: Warning: The this-pointer is stored as context in the signal.
            This pointer gets copied, when the signal is copied (or downcasted)
            to other signals, and thus lives on, even when this actual object
            gets deleted. The cable function acesses that pointer and does not
            know, whether this original is still alive or has been deleted any
            time ago.

            Therefore, the object of type TDangerousCableToSlot may not be
            deleted as long as there is a signal that uses this cable (i.e.
            is a copy of this cable).

            That is the reason for "dangerous" in the class name.
        */
    {
        printf ("Constructor of TDangerousCableToSlot (const variant) is called.\n");
    }

    TDangerousCableToSlot(tReceiver &receiver, memberFunctionPointer_nonconst slot)
    :
        m_receiver(&receiver),
        m_slot_const(0),
        m_slot_nonconst(slot),
        CSignal(cableFunction_nonconst, this)
        /*
            TODO: Warning: The this-pointer is stored as context in the signal.
            This pointer gets copied, when the signal is copied (or downcasted)
            to other signals, and thus lives on, even when this actual object
            gets deleted. The cable function acesses that pointer and does not
            know, whether this original is still alive or has been deleted any
            time ago.

            Therefore, the object of type TDangerousCableToSlot may not be
            deleted as long as there is a signal that uses this cable (i.e.
            is a copy of this cable).

            That is the reason for "dangerous" in the class name.
        */
    {
        printf ("Constructor of TDangerousCableToSlot (nonconst variant) is called.\n");
    }

    TDangerousCableToSlot(tReceiver &receiver, memberFunctionPointer_static slot)
    :
        m_receiver(0),
        m_slot_const(0),
        m_slot_nonconst(0),
        CSignal(cableFunction_static, slot)
    {
        printf ("Constructor of TDangerousCableToSlot (static variant) is called.\n");
    }

    private:

    const tReceiver *m_receiver;
    memberFunctionPointer_const m_slot_const;
    memberFunctionPointer_nonconst m_slot_nonconst;

    static void cableFunction_const (void *context, const typename tSignature::CMessage &msg)
    {
        TDangerousCableToSlot *this_ = static_cast<TDangerousCableToSlot *>(context);
        tSlot::callSlot (*(this_->m_receiver), this_->m_slot_const, msg);
    }

    static void cableFunction_nonconst (void *context, const typename tSignature::CMessage &msg)
    {
        TDangerousCableToSlot *this_ = static_cast<TDangerousCableToSlot *>(context);

        tReceiver *receiver = const_cast<tReceiver *>(this_->m_receiver);
        tSlot::callSlot (*receiver, this_->m_slot_nonconst, msg);
    }

    static void cableFunction_static (void *context, const typename tSignature::CMessage &msg)
    {
        memberFunctionPointer_static slot_static = static_cast<memberFunctionPointer_static>(context);
        tSlot::callSlot (slot_static, msg);
    }
};


/*!
    This \c CManualDemonstrator demonstrates how to use
    the Sigslot concept and ist library to create own
    applications without much effort.

    The generation of connections involves some dynamic
    memory allocation and stores more data than necessary
    compared to automatically generated code (e.g. some
    member function pointers).
*/

class CManualDemonstrator
{
    private:

    CInput      m_input;
    CMultiply   m_mult;
    CAverage    m_average1;
    CAverage    m_average2;
    CPrint      m_print;

    public:

    CManualDemonstrator();
    void processAValue(double value);
};

CManualDemonstrator::CManualDemonstrator()
{
    /*
        Create the connections...
    */

//    TCableToSlot (m_input.SIGNAL_newValue, m_print.SLOT_print);
    m_input.SIGNAL_newValue = TIncorrectCableToSlot<NSigslot::TSlot1<tValueSignature, CPrint> > (m_print, CPrint::SLOT_print);
    m_input.SIGNAL_newValue = *new TDangerousCableToSlot<NSigslot::TSlot1<tValueSignature, CPrint> > (m_print, CPrint::SLOT_print);
}

void CManualDemonstrator::processAValue(double value)
{
    m_input.processInput (value);
}




// ******************************************************************
// ******************************************************************
// ******************************************************************

/*
    This section demonstrates all the possible cables and receivers.
*/

typedef NSigslot::TSignature2<int, const char *> tDemoSig;

class CSender1
{
    public:

    tDemoSig::CSignal SIGNAL_sig;
};

class CReceiver1
{
    public:

    void operator()(int i, const char *s) const
    {
        printf ("const operator() of Receiver1 %p is called (i=%d, s=%s).\n", this, i, s);
    }

    void operator()(int i, const char *s)
    {
        printf ("non-const operator() of Receiver1 %p is called (i=%d, s=%s).\n", this, i, s);
    }

    void operator()(const tDemoSig::CMessage &msg) const
    {
        printf ("const operator() (msg-by-ref-version) of Receiver1 %p is called (i=%d, s=%s).\n", this, msg.m_par1, msg.m_par2);
    }

    void operator()(const tDemoSig::CMessage &msg)
    {
        printf ("non-const operator() (msg-by-ref-version) of Receiver1 %p is called (i=%d, s=%s).\n", this, msg.m_par1, msg.m_par2);
    }

    // Note: operator() may not be static.

    void SLOT_const (int i, const char *s) const
    {
        printf ("SLOT_const of Receiver1 %p is called (i=%d, s=%s).\n", this, i, s);
    }

    void SLOT_nonconst (int i, const char *s)
    {
        printf ("SLOT_nonconst of Receiver1 %p is called (i=%d, s=%s).\n", this, i, s);
    }

    static void SLOT_static (int i, const char *s)
    {
        printf ("SLOT_static of class Receiver1 is called (i=%d, s=%s).\n", i, s);
    }
};

class CReceiver2
{
    public:

    void operator()(tDemoSig::CMessage msg) const
    {
        printf ("const operator() (msg by value of Receiver1 %p is called (i=%d, s=%s).\n", this, msg.m_par1, msg.m_par2);
    }

    void operator()(tDemoSig::CMessage msg)
    {
        printf ("non-const operator() (msg by value) of Receiver1 %p is called (i=%d, s=%s).\n", this, msg.m_par1, msg.m_par2);
    }
};

static void SLOT_static_function (int i, const char *s)
{
    printf ("SLOT_static_function in global namespace is called (i=%d, s=%s).\n", i, s);
}

void cableDemo ()
{
    CSender1 s1;
    CReceiver1 r1;
    const CReceiver1 r2;
    CReceiver2 r3;
    const CReceiver2 r4;

    s1.SIGNAL_sig(666, "not connected");

    typedef NSigslot::TSlot<tDemoSig, CReceiver1> tSlotOfCReceiver1;
    s1.SIGNAL_sig = NSigslot::TCableToMessageOperator<tSlotOfCReceiver1>(r1);
    s1.SIGNAL_sig(1, "msg-by-ref-operator (nonconst)");

    s1.SIGNAL_sig = NSigslot::TCableToMessageOperator<tSlotOfCReceiver1>(r2);
    s1.SIGNAL_sig(2, "msg-by-ref-operator (const)");

    typedef NSigslot::TSlot<tDemoSig, CReceiver2> tSlotOfCReceiver2;
    s1.SIGNAL_sig = NSigslot::TCableToMessageOperator<tSlotOfCReceiver2>(r3);
    s1.SIGNAL_sig(3, "msg-by-value-operator (nonconst)");

    s1.SIGNAL_sig = NSigslot::TCableToMessageOperator<tSlotOfCReceiver2>(r4);
    s1.SIGNAL_sig(4, "msg-by-value-operator (const)");

    typedef NSigslot::TSlot2<tDemoSig, CReceiver1> tSlot2OfCReceiver1;
    s1.SIGNAL_sig = NSigslot::TCableToOperator<tSlot2OfCReceiver1>(r1);
    s1.SIGNAL_sig(5, "normal-operator (nonconst)");

    s1.SIGNAL_sig = NSigslot::TCableToOperator<tSlot2OfCReceiver1>(r2);
    s1.SIGNAL_sig(6, "normal-operator (const)");

    s1.SIGNAL_sig = TIncorrectCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_nonconst);
    s1.SIGNAL_sig(7, "normal slot r1 (nonconst)");
    
    s1.SIGNAL_sig = TIncorrectCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_const);
    s1.SIGNAL_sig(8, "normal slot r1 (const)");
    
    s1.SIGNAL_sig = TIncorrectCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_static);
    s1.SIGNAL_sig(9, "normal slot r1 (static)");
    
    s1.SIGNAL_sig = *new TDangerousCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_nonconst);
    s1.SIGNAL_sig(10, "normal slot r1 (nonconst)");
    
    s1.SIGNAL_sig = *new TDangerousCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_const);
    s1.SIGNAL_sig(11, "normal slot r1 (const)");
    
    s1.SIGNAL_sig = *new TDangerousCableToSlot<tSlot2OfCReceiver1>(r1, r1.SLOT_static);
    s1.SIGNAL_sig(12, "normal slot r1 (static)");
    
    s1.SIGNAL_sig = TProblematicCableToSlot<tSlot2OfCReceiver1, tSlot2OfCReceiver1::memberFunctionPointer_nonconst>(r1, r1.SLOT_nonconst);
    s1.SIGNAL_sig(13, "normal slot r1 (nonconst)");
    
    s1.SIGNAL_sig = TProblematicCableToSlot<tSlot2OfCReceiver1, tSlot2OfCReceiver1::memberFunctionPointer_const>(r1, r1.SLOT_const);
    s1.SIGNAL_sig(14, "normal slot r1 (const)");
    
    s1.SIGNAL_sig = TProblematicCableToSlot<tSlot2OfCReceiver1, tSlot2OfCReceiver1::memberFunctionPointer_static>(r1, r1.SLOT_static);
    s1.SIGNAL_sig(15, "normal slot r1 (static)");
    
    s1.SIGNAL_sig = tDemoSig::CSignal();
    s1.SIGNAL_sig(666, "no longer connected");
}


// ******************************************************************
// ******************************************************************
// ******************************************************************

class CRandom
{
    public:

    void SLOT_rollDice();
    tValueSignature::CSignal SIGNAL_value;
};

void CRandom::SLOT_rollDice()
{
    double randomValue = 2.7;
    NBase::CThread::sleep (200);  // simulate a calculation time of 0.2 seconds...
    SIGNAL_value (randomValue);
}


class CThreadTest
{
    private:

    NSigslot::CThread   m_thread;
    CRandom             m_random;
    CPrint              m_print;

    public:

    CThreadTest();
    void start();
    void stop();

    private:

    bool m_threadIsActive;

    static void cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg);
    static void cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_random_value (void *context, const tValueSignature::CMessage &msg);
};


void CThreadTest::cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg)
{
    // LOCK

    CThreadTest *this_ = static_cast<CThreadTest *>(context);
    this_->m_threadIsActive = msg.m_par1;
}

void CThreadTest::cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CThreadTest *this_ = static_cast<CThreadTest *>(context);
    this_->m_random.SLOT_rollDice();
}

void CThreadTest::cable_random_value (void *context, const tValueSignature::CMessage &msg)
{
    CThreadTest *this_ = static_cast<CThreadTest *>(context);
    this_->m_print.SLOT_print(msg.m_par1);
}

CThreadTest::CThreadTest()
:
    m_threadIsActive(false)
{
    /*
        Create the connections...
    */

    m_thread.INIT_active = NSigslot::TSignature1<bool>::CSignal(cable_thread_init_active, this);
    m_thread.SIGNAL_thread = NSigslot::TSignature0::CSignal(cable_thread_thread, this);
    m_random.SIGNAL_value = tValueSignature::CSignal(cable_random_value, this);
}

void CThreadTest::start()
{
    m_thread.SLOT_start();
}

void CThreadTest::stop()
{
    while (m_threadIsActive)    // TODO: Lock access to m_threadIsActive !
    {
        m_thread.INIT_shutdown();
    }
}


// ******************************************************************
// ******************************************************************
// ******************************************************************


#include "Sigslot/CWait.hpp"
#include "Sigslot/TQueue.hpp"
#include <deque>


class CProcessValue
{
    public:

    void SLOT_calculate(double value);
    tValueSignature::CSignal SIGNAL_result;
};

void CProcessValue::SLOT_calculate(double value)
{
    printf ("processValue: calculate(%f) called...\n", value);
    NBase::CThread::sleep (500);  // simulate a calculation time of half a second...
    SIGNAL_result (value);
}


// ******************************************************************

/*!
    The composiste component (or application) CQueueTest1 uses a
    CThread, CWait and CQueue to implement a asynchronous connection
    between a CRandom and a CProcessValue component.

    There is somewhere a graphical diagram that shows the connections.
    At least, Kurt has a piece of paper (2005-08-25) with some boxes
    and lines.

    This implementation is not the typical implementation. It works
    only for signatures with exactly one parameter.

    See \c CQueueTest2 for exactly the same example. That implementation
    shows the generic way that works with 0 or 2 or more parameters
    as well (even when it is shown, it _is_ the same example, with
    one parameter).
*/
class CQueueTest1
{
    private:

    NSigslot::CThread   m_thread;
    NSigslot::CWait     m_wait;
    NSigslot::TQueue<double, std::deque<double> >
                        m_queue;
    CRandom             m_random;
    CProcessValue       m_processValue;
    CPrint              m_print;

    public:

    CQueueTest1();
    void start();
    void runApplication();
    void stop();

    private:

    bool m_threadIsActive;

    static void cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg);
    static void cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_random_value (void *context, const tValueSignature::CMessage &msg);
    static void cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_message (void *context, const tValueSignature::CMessage &msg);
    static void cable_processValue_result (void *context, const tValueSignature::CMessage &msg);
};

void CQueueTest1::cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg)
{
    // LOCK

    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_threadIsActive = msg.m_par1;
}

void CQueueTest1::cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_wait.SLOT_work();
}

void CQueueTest1::cable_random_value (void *context, const tValueSignature::CMessage &msg)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_queue(msg.m_par1);
}

void CQueueTest1::cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_queue.SLOT_read();
}

void CQueueTest1::cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_wait.SLOT_wakeup();
}

void CQueueTest1::cable_queue_message (void *context, const tValueSignature::CMessage &msg)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_processValue.SLOT_calculate(msg.m_par1);
}

void CQueueTest1::cable_processValue_result (void *context, const tValueSignature::CMessage &msg)
{
    CQueueTest1 *this_ = static_cast<CQueueTest1 *>(context);
    this_->m_print.SLOT_print(msg.m_par1);
}

CQueueTest1::CQueueTest1()
:
    m_threadIsActive(false)
{
    /*
        Create the connections...
    */

    m_thread.INIT_active = NSigslot::TSignature1<bool>::CSignal(cable_thread_init_active, this);
    m_thread.SIGNAL_thread = NSigslot::TSignature0::CSignal(cable_thread_thread, this);
    m_random.SIGNAL_value = tValueSignature::CSignal(cable_random_value, this);
    m_wait.SIGNAL_work = NSigslot::TSignature0::CSignal(cable_wait_work, this);
    m_queue.SIGNAL_notify = NSigslot::TSignature0::CSignal(cable_queue_notify, this);
    m_queue.SIGNAL_message = tValueSignature::CSignal(cable_queue_message, this);
    m_processValue.SIGNAL_result = tValueSignature::CSignal(cable_processValue_result, this);
}

void CQueueTest1::start()
{
    m_thread.SLOT_start();
}

void CQueueTest1::runApplication()
{
    for (int i = 0; i<6; i++)
    {
        printf ("Roll Dice #%d\n", i);
        m_random.SLOT_rollDice();
    }
}

void CQueueTest1::stop()
{
    while (m_threadIsActive)    // TODO: Lock access to m_threadIsActive !
    {
        m_thread.INIT_shutdown();
        m_wait.INIT_shutdown();
    }
}


// ******************************************************************

/*!
    The composiste component (or application) CQueueTest2 uses a
    CThread, CWait and CQueue to implement a asynchronous connection
    between a CRandom and a CProcessValue component.

    There is somewhere a graphical diagram that shows the connections.
    At least, Kurt has a piece of paper (2005-08-25) with some boxes
    and lines.

    The template NSigslot::TQueue is instantiated with the CMessage,
    that is the structure that holds all the parameters of the queued
    signal.
*/
class CQueueTest2
{
    private:

    NSigslot::CThread   m_thread;
    NSigslot::CWait     m_wait;
    NSigslot::TQueue<tValueSignature::CMessage, std::deque<tValueSignature::CMessage> >
                        m_queue;
    CRandom             m_random;
    CProcessValue       m_processValue;
    CPrint              m_print;

    public:

    CQueueTest2();
    void start();
    void runApplication();
    void stop();

    private:

    bool m_threadIsActive;

    static void cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg);
    static void cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_random_value (void *context, const tValueSignature::CMessage &msg);
    static void cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_message (void *context, const NSigslot::TSignature1<tValueSignature::CMessage>::CMessage &msg);
    static void cable_processValue_result (void *context, const tValueSignature::CMessage &msg);
};

void CQueueTest2::cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg)
{
    // LOCK

    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_threadIsActive = msg.m_par1;
}

void CQueueTest2::cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_wait.SLOT_work();
}

void CQueueTest2::cable_random_value (void *context, const tValueSignature::CMessage &msg)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_queue(msg);
}

void CQueueTest2::cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_queue.SLOT_read();
}

void CQueueTest2::cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_wait.SLOT_wakeup();
}

void CQueueTest2::cable_queue_message (void *context, const NSigslot::TSignature1<tValueSignature::CMessage>::CMessage &msg)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_processValue.SLOT_calculate(msg.m_par1.m_par1);
}

void CQueueTest2::cable_processValue_result (void *context, const tValueSignature::CMessage &msg)
{
    CQueueTest2 *this_ = static_cast<CQueueTest2 *>(context);
    this_->m_print.SLOT_print(msg.m_par1);
}

CQueueTest2::CQueueTest2()
:
    m_threadIsActive(false)
{
    /*
        Create the connections...
    */

    m_thread.INIT_active = NSigslot::TSignature1<bool>::CSignal(cable_thread_init_active, this);
    m_thread.SIGNAL_thread = NSigslot::TSignature0::CSignal(cable_thread_thread, this);
    m_random.SIGNAL_value = tValueSignature::CSignal(cable_random_value, this);
    m_wait.SIGNAL_work = NSigslot::TSignature0::CSignal(cable_wait_work, this);
    m_queue.SIGNAL_notify = NSigslot::TSignature0::CSignal(cable_queue_notify, this);
    m_queue.SIGNAL_message = NSigslot::TSignature1<tValueSignature::CMessage>::CSignal(cable_queue_message, this);
    m_processValue.SIGNAL_result = tValueSignature::CSignal(cable_processValue_result, this);
}

void CQueueTest2::start()
{
    m_thread.SLOT_start();
}

void CQueueTest2::runApplication()
{
    for (int i = 0; i<6; i++)
    {
        printf ("Roll Dice #%d\n", i);
        m_random.SLOT_rollDice();
    }
}

void CQueueTest2::stop()
{
    while (m_threadIsActive)    // TODO: Lock access to m_threadIsActive !
    {
        m_thread.INIT_shutdown();
        m_wait.INIT_shutdown();
    }
}


// ******************************************************************
// ******************************************************************
// ******************************************************************

#include "Sigslot/TMessageUnion.hpp"

void testMessageUnion ()
{
    typedef NSigslot::TMessageUnion0 t_mu0;
    typedef NSigslot::TMessageUnion1<int> t_mu1;
    typedef NSigslot::TMessageUnion3<int, double, int> t_mu3;

    t_mu0 mu0;
    t_mu1 mu1;
    t_mu3 mu3;

    mu1.setValue1 (42);
    assert (mu1.getValue1() == 42);

    mu3.setValue3(23);
    assert (mu3.getValue3() == 23);
//    mu3.getValue1(); // fails!

    printf ("Simple tests OK. Most of all, it compiles!\n");
}


// ******************************************************************
// ******************************************************************
// ******************************************************************


/*!
    CMachine simulates a component with three slots, that all take
    a long time to process and have to be called synchronized (that
    is, one one of these three slots may be called at a time).
*/

class CMachine
{
    public:

    void SLOT_nr1(double value);
    void SLOT_nr2(std::string msg, int i);
    void SLOT_nr3(double value);

    CMachine()
    :
        m_counter(0)
    {
    }

    private:

    int m_counter;
};

void CMachine::SLOT_nr1 (double value)
{
    printf ("CMachine(%p), SLOT_nr1 is called with value %f.\n", this, value);
    m_counter++;
    assert (m_counter == 1);
    NBase::CThread::sleep (500);  // simulate a calculation time of half a second...
    m_counter--;
    assert (m_counter == 0);
    printf ("CMachine(%p), SLOT_nr1 returns now.\n", this);
}

void CMachine::SLOT_nr2 (std::string msg, int i)
{
    printf ("CMachine(%p), SLOT_nr2 is called with value '%s' (i=%d).\n", this, msg.c_str(), i);
    m_counter++;
    assert (m_counter == 1);
    NBase::CThread::sleep (500);  // simulate a calculation time of half a second...
    m_counter--;
    assert (m_counter == 0);
    printf ("CMachine(%p), SLOT_nr2 returns now.\n", this);
}

void CMachine::SLOT_nr3 (double value)
{
    printf ("CMachine(%p), SLOT_nr3 is called with value %f.\n", this, value);
    m_counter++;
    assert (m_counter == 1);
    NBase::CThread::sleep (500);  // simulate a calculation time of half a second...
    m_counter--;
    assert (m_counter == 0);
    printf ("CMachine(%p), SLOT_nr3 returns now.\n", this);
}



/*!
    The composiste component (or application) CQueueTest3 uses a
    CThread, CWait and CQueue to implement a asynchronous connection.
    This time, CQueue is instantialted with a TMessageUnion.
*/

class CQueueTest3
{
    private:

    typedef NSigslot::TSignature1<double>           tSig1;
    typedef NSigslot::TSignature2<std::string, int> tSig2;
    typedef NSigslot::TSignature1<double>           tSig3;  // identical to tSig1 to check this complication

    typedef NSigslot::TSignature1<tSig1::CMessage>   tSig1Message;
    typedef NSigslot::TSignature1<tSig2::CMessage>   tSig2Message;
    typedef NSigslot::TSignature1<tSig3::CMessage>   tSig3Message;

    typedef NSigslot::TMessageUnion3<tSig1::CMessage, tSig2::CMessage, tSig3::CMessage> tMsgUnion;
//  typedef NSigslot::TMessageUnion3<int, std::string, double> tMsgUnion;
    typedef NSigslot::TSignature1<tMsgUnion> tSignatureMessageUnion;

    NSigslot::CThread   m_thread;
    NSigslot::CWait     m_wait;
    NSigslot::TQueue<tMsgUnion, std::deque<tMsgUnion> >
                        m_queue;
    NSigslot::TEnpatcher3<tMsgUnion> m_enpatcher;
    NSigslot::TDispatcher3<tMsgUnion> m_dispatcher;
    CMachine            m_machine;

    public:

    CQueueTest3();
    void start();
    void runApplication();
    void stop();

    private:

    bool m_threadIsActive;

    static void cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg);
    static void cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &msg);
    static void cable_queue_message (void *context, const NSigslot::TSignature1<tMsgUnion>::CMessage &msg);
    static void cable_enpatcher_signal (void *context, const tSignatureMessageUnion::CMessage &msg);
    static void cable_dispatcher_nr1 (void *context, const tSig1Message::CMessage &msg);
    static void cable_dispatcher_nr2 (void *context, const tSig2Message::CMessage &msg);
    static void cable_dispatcher_nr3 (void *context, const tSig3Message::CMessage &msg);
};

void CQueueTest3::cable_thread_init_active (void *context, const NSigslot::TSignature1<bool>::CMessage &msg)
{
    // LOCK

    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_threadIsActive = msg.m_par1;
}

void CQueueTest3::cable_thread_thread (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_wait.SLOT_work();
}

void CQueueTest3::cable_wait_work (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_queue.SLOT_read();
}

void CQueueTest3::cable_queue_notify (void *context, const NSigslot::TSignature0::CMessage &/*msg*/)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_wait.SLOT_wakeup();
}

void CQueueTest3::cable_queue_message (void *context, const NSigslot::TSignature1<tMsgUnion>::CMessage &msg)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_dispatcher(msg.m_par1);
}

void CQueueTest3::cable_enpatcher_signal (void *context, const tSignatureMessageUnion::CMessage &msg)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_queue(msg.m_par1);
}

void CQueueTest3::cable_dispatcher_nr1 (void *context, const tSig1Message::CMessage &msg)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_machine.SLOT_nr1(msg.m_par1.m_par1);
}

void CQueueTest3::cable_dispatcher_nr2 (void *context, const tSig2Message::CMessage &msg)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_machine.SLOT_nr2(msg.m_par1.m_par1, msg.m_par1.m_par2);
}

void CQueueTest3::cable_dispatcher_nr3 (void *context, const tSig3Message::CMessage &msg)
{
    CQueueTest3 *this_ = static_cast<CQueueTest3 *>(context);
    this_->m_machine.SLOT_nr3(msg.m_par1.m_par1);
}

CQueueTest3::CQueueTest3()
:
    m_threadIsActive(false)
{
    /*
        Create the connections...
    */

    m_thread.INIT_active = NSigslot::TSignature1<bool>::CSignal(cable_thread_init_active, this);
    m_thread.SIGNAL_thread = NSigslot::TSignature0::CSignal(cable_thread_thread, this);
    m_wait.SIGNAL_work = NSigslot::TSignature0::CSignal(cable_wait_work, this);
    m_queue.SIGNAL_notify = NSigslot::TSignature0::CSignal(cable_queue_notify, this);
    m_queue.SIGNAL_message = NSigslot::TSignature1<tMsgUnion>::CSignal(cable_queue_message, this);
    m_enpatcher.SIGNAL = tSignatureMessageUnion::CSignal(cable_enpatcher_signal, this);
    m_dispatcher.SIGNAL_nr1 = tSig1Message::CSignal(cable_dispatcher_nr1, this);
    m_dispatcher.SIGNAL_nr2 = tSig2Message::CSignal(cable_dispatcher_nr2, this);
    m_dispatcher.SIGNAL_nr3 = tSig3Message::CSignal(cable_dispatcher_nr3, this);
}

void CQueueTest3::start()
{
    m_thread.SLOT_start();
}

void CQueueTest3::runApplication()
{
    for (int i = 0; i<6; i++)
    {
        printf ("Roll Dice #%d\n", i);
        m_enpatcher.SLOT_nr1(tSig1::CMessage(42.1));
        m_enpatcher.SLOT_nr2(tSig2::CMessage("some text", 23));
// NBase::CThread::sleep(0);
        m_enpatcher.SLOT_nr3(tSig3::CMessage(23.7));
        m_enpatcher.SLOT_nr1(tSig1::CMessage(42.2));
    }

//    m_machine.SLOT_nr2 ("calling not synchronized from another thread. That will crash!", -1);
}

void CQueueTest3::stop()
{
    while (m_threadIsActive)    // TODO: Lock access to m_threadIsActive !
    {
        m_thread.INIT_shutdown();
        m_wait.INIT_shutdown();
    }
}


// ******************************************************************
// ******************************************************************
// ******************************************************************

int main(int argc, char* argv[])
{
	printf("Simple Demo for the Sigslot Concept\n");

    printf ("Creating the components...\n");

    CCalculator calculator1;
    CCalculator calculator2;

//    printf ("Creating the connections...\n");

    calculator1.createConnections();

    printf ("Fire some values to calculator 1...\n");

    calculator1.processAValue (23);
    calculator1.processAValue (42);
    calculator1.processAValue (-1);

    printf ("Fire some values to calculator 2...\n");

    calculator2.processAValue (23);
    calculator2.processAValue (42);
    calculator2.processAValue (-1);

    printf ("\n");
    printf ("And now for the manual demonstrator.\n");
    printf ("\n");

    CManualDemonstrator demo;
    demo.processAValue(17);
    demo.processAValue(4);

    printf ("\n");
    printf ("And now for a demo of different cables.\n");
    printf ("\n");

    cableDemo();

    printf ("\n");
    printf ("And now for a test for CThread components.\n");
    printf ("\n");

    CThreadTest threadTest;
    threadTest.start();
    NBase::CThread::sleep (5);     // sleep for 5 milliseconds
//    NBase::CThread::sleep (1000);  // sleep for a second
    threadTest.stop();

#if 0
    printf ("\n");
    printf ("And now for a test for CQueue components.\n");
    printf ("\n");

    CQueueTest1 queueTest1;
    queueTest1.start();
    printf ("Enter some values into the queue.\n");
    queueTest1.runApplication();
    printf ("Input is stored in the queue, and some of it already executed.\n");
    NBase::CThread::sleep (2000);  // sleep for 2 further seconds
    printf ("Now stop the component, even when the queue is not yet empty.\n");
    queueTest1.stop();
#endif

#if 0
    printf ("\n");
    printf ("And now for a second test for CQueue components.\n");
    printf ("\n");

    CQueueTest2 queueTest2;
    queueTest2.start();
    printf ("Enter some values into the queue.\n");
    queueTest2.runApplication();
    printf ("Input is stored in the queue, and some of it already executed.\n");
    NBase::CThread::sleep (2000);  // sleep for 2 further seconds
    printf ("Now stop the component, even when the queue is not yet empty.\n");
    queueTest2.stop();
#endif

    printf ("\n");
    printf ("And now comes a boring test for TMessageUnion.\n");
    printf ("\n");

    testMessageUnion ();

#if 1
    printf ("\n");
    printf ("And now for a third test for CQueue components using TMessageUnion.\n");
    printf ("\n");

    CQueueTest3 queueTest3;
    queueTest3.start();
    printf ("Enter some values into the queue.\n");
    queueTest3.runApplication();
    printf ("Input is stored in the queue, and some of it already executed.\n");
//    NBase::CThread::sleep (2000);  // sleep for 2 further seconds
//    NBase::CThread::sleep (2000);  // sleep for 2 further seconds
    printf ("Now stop the component, even when the queue is not yet empty.\n");
    queueTest3.stop();
#endif

    printf ("\n");
    printf ("And now comes the code used for Sigslot presentations.\n");
    printf ("\n");

    extern void run_sigslot_presentation();
    run_sigslot_presentation();

    printf ("\n");
    printf ("That's it for the first\n");
    printf ("\n");

	return 0;
}

